@@ -2,8 +2,8 @@
 
   class.c -
 
-  $Author: matz $
-  $Date: 2003/08/04 08:31:20 $
+  $Author: nobu $
+  $Date: 2003/12/22 12:15:34 $
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -14,7 +14,6 @@
 #include "rubysig.h"
 #include "node.h"
 #include "st.h"
-#include "version.h"
 #include <ctype.h>
 
 extern st_table *rb_class_tbl;
@@ -60,39 +59,40 @@ clone_method(mid, body, tbl)
 }
 
 VALUE
-rb_mod_clone(module)
-    VALUE module;
+rb_mod_init_copy(clone, orig)
+    VALUE clone, orig;
 {
-    NEWOBJ(clone, struct RClass);
-    CLONESETUP(clone, module);
-
-    RCLASS(clone)->super = RCLASS(module)->super;
-    if (RCLASS(module)->iv_tbl) {
+    rb_obj_init_copy(clone, orig);
+    if (!FL_TEST(CLASS_OF(clone), FL_SINGLETON)) {
+	RBASIC(clone)->klass = rb_singleton_class_clone(orig);
+    }
+    RCLASS(clone)->super = RCLASS(orig)->super;
+    if (RCLASS(orig)->iv_tbl) {
 	ID id;
 
-	RCLASS(clone)->iv_tbl = st_copy(RCLASS(module)->iv_tbl);
+	RCLASS(clone)->iv_tbl = st_copy(RCLASS(orig)->iv_tbl);
 	id = rb_intern("__classpath__");
 	st_delete(RCLASS(clone)->iv_tbl, (st_data_t*)&id, 0);
 	id = rb_intern("__classid__");
 	st_delete(RCLASS(clone)->iv_tbl, (st_data_t*)&id, 0);
     }
-    if (RCLASS(module)->m_tbl) {
+    if (RCLASS(orig)->m_tbl) {
 	RCLASS(clone)->m_tbl = st_init_numtable();
-	st_foreach(RCLASS(module)->m_tbl, clone_method,
+	st_foreach(RCLASS(orig)->m_tbl, clone_method,
 	  (st_data_t)RCLASS(clone)->m_tbl);
     }
 
-    return (VALUE)clone;
+    return clone;
 }
 
 VALUE
-rb_mod_dup(mod)
-    VALUE mod;
+rb_class_init_copy(clone, orig)
+    VALUE clone, orig;
 {
-    VALUE dup = rb_mod_clone(mod);
-
-    RBASIC(dup)->flags = RBASIC(mod)->flags & (T_MASK|FL_TAINT|FL_SINGLETON);
-    return dup;
+    if (RCLASS(clone)->super != 0) {
+	rb_raise(rb_eTypeError, "already initialized class");
+    }
+    return rb_mod_init_copy(clone, orig);
 }
 
 VALUE
@@ -181,6 +181,19 @@ rb_define_class_id(id, super)
     return klass;
 }
 
+void
+rb_check_inheritable(super)
+    VALUE super;
+{
+    if (TYPE(super) != T_CLASS) {
+	rb_raise(rb_eTypeError, "superclass must be a Class (%s given)",
+		 rb_obj_classname(super));
+    }
+    if (RBASIC(super)->flags & FL_SINGLETON) {
+	rb_raise(rb_eTypeError, "can't make subclass of virtual class");
+    }
+}
+
 VALUE
 rb_class_inherited(super, klass)
     VALUE super, klass;
@@ -434,8 +447,7 @@ VALUE
 rb_mod_ancestors(mod)
     VALUE mod;
 {
-    VALUE ary = rb_ary_new();
-    VALUE p;
+    VALUE p, ary = rb_ary_new();
 
     for (p = mod; p; p = RCLASS(p)->super) {
 	if (FL_TEST(p, FL_SINGLETON))
@@ -542,12 +554,7 @@ class_instance_method_list(argc, argv, mod, func)
     st_table *list;
 
     if (argc == 0) {
-#if RUBY_VERSION_CODE < 181
-	rb_warn("%s: parameter will default to 'true' as of 1.8.1", rb_id2name(rb_frame_last_func()));
-	recur = Qfalse;
-#else
 	recur = Qtrue;
-#endif
     }
     else {
 	VALUE r;
@@ -616,11 +623,7 @@ rb_obj_singleton_methods(argc, argv, obj)
 
     rb_scan_args(argc, argv, "01", &recur);
     if (argc == 0) {
-#if RUBY_VERSION_CODE < 181
-	rb_warn("singleton_methods: parameter will default to 'true' as of 1.8.1");
-#else
 	recur = Qtrue;
-#endif
     }
     klass = CLASS_OF(obj);
     list = st_init_numtable();
